#ifndef __RENDCLIPFRUST__
#define __RENDCLIPFRUST__

#include "rendBaseTypes.h"
#include "rendVect.h"

// ClipByW_ge()		: clips a polygon by one of its vertex coordinates so this coord. is greater or equal to -W
//		VECTOR_T						: one of Vector* types, must be of R4 or greater
//		- coord,	int32				: index of the coordinate to clip by,		input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByW_ge(int32 coord,
					   const VECTOR_T *poly[],
					   uint32 &start,
					   uint32 &size,
					   VECTOR_T buffer[],
					   uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = (*poly[currVert])[coord] + (*poly[currVert])[3];
		float distPrev = (*poly[prevVert])[coord] + (*poly[prevVert])[3];

		if (distCurr < 0) // is current vertex outside?
		{
			if (distPrev > 0) // is previous vertex strictly inside?
			{
				Interpolate< VECTOR_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr > 0 &&	// is current vertex strictly inside and
				distPrev < 0)	// is previous vertex outside?
			{
				Interpolate< VECTOR_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByW_le()		: clips a polygon by one of its vertex coordinates so this coord. is less or equal to W
//		VECTOR_T						: one of Vector* types, must be of R4 or greater
//		- coord,	int32				: index of the coordinate to clip,			input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByW_le(int32 coord,
					   const VECTOR_T *poly[],
					   uint32 &start,
					   uint32 &size,
					   VECTOR_T buffer[],
					   uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = (*poly[currVert])[coord] - (*poly[currVert])[3];
		float distPrev = (*poly[prevVert])[coord] - (*poly[prevVert])[3];

		if (distCurr > 0) // is current vertex outside?
		{
			if (distPrev < 0) // is previous vertex strictly inside?
			{
				Interpolate< VECTOR_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr < 0 &&	// is current vertex strictly inside and
				distPrev > 0)	// is previous vertex outside?
			{
				Interpolate< VECTOR_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}


// ClipByFrustum()	: clips a polygon by the view frustum
//		VECTOR_T						: one of Vector* types, must be of R4 or greater
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- size would be 0 if clipping discarded all vertices
template < typename VECTOR_T >

inline void ClipByFrustum(const VECTOR_T *poly[],
						  uint32 &start,
						  uint32 &size,
						  VECTOR_T buffer[],
						  uint32 &nbuff)
{
	// -W <= X <= W
	ClipByW_ge< VECTOR_T >(0, poly, start, size, buffer, nbuff);
	ClipByW_le< VECTOR_T >(0, poly, start, size, buffer, nbuff);

	// -W <= Y <= W
	ClipByW_ge< VECTOR_T >(1, poly, start, size, buffer, nbuff);
	ClipByW_le< VECTOR_T >(1, poly, start, size, buffer, nbuff);

	// -W <= Z <= W
	ClipByW_ge< VECTOR_T >(2, poly, start, size, buffer, nbuff);
	ClipByW_le< VECTOR_T >(2, poly, start, size, buffer, nbuff);
}

#endif // __RENDCLIPFRUST__
