#ifndef __RENDCLIPRECT__
#define __RENDCLIPRECT__

#include "rendBaseTypes.h"
#include "rendHelper.h"

// IntersectCol()		: finds the intersection between a line segment and a vertical line
//		VECTOR_T							: one of Vector* types; must be of R2 or greater
//		- p0,			const <VECTOR_T> *	: line segment 1st vertex,	input
//		- p1,			const <VECTOR_T> *	: line segment 2nd vertex,	input
//		- x,			float				: column to intersect with,	input
//		- p_intersect,	<VECTOR_T> *		: point of intersection,	output
// returns
//		nil
// note
//		- line segment must be non-vertical
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void IntersectCol(const VECTOR_T *p0,
						 const VECTOR_T *p1,
						 float x,
						 VECTOR_T *p_intersect)
{
	VECTOR_T d;
	
	for (int comp = 0; comp < VecSize< VECTOR_T >(); comp++)
		d[comp] = (*p1)[comp] - (*p0)[comp];

	const float ox = x - (*(const Vertex2 *)p0).x;

	(*p_intersect)[0] = x;
	
	for (int comp = 1; comp < VecSize< VECTOR_T >(); comp++)
		(*p_intersect)[comp] = ox * d[comp] / d[0] + (*p0)[comp];
}

// IntersectRow()		: finds the intersection between a line segment and a horizontal line
//		VECTOR_T							: one of Vector* types; must be of R2 or greater
//		- p0,			const <VECTOR_T> *	: line segment 1st vertex,	input
//		- p1,			const <VECTOR_T> *	: line segment 2nd vertex,	input
//		- y,			float				: row to intersect with,	input
//		- p_intersect,	<VECTOR_T> *		: point of intersection,	output
// returns
//		nil
// note
//		- line segment must be non-horizontal
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void IntersectRow(const VECTOR_T *p0,
						 const VECTOR_T *p1,
						 float y,
						 VECTOR_T *p_intersect)
{
	VECTOR_T d;

	for (int comp = 0; comp < VecSize< VECTOR_T >(); comp++)
		d[comp] = (*p1)[comp] - (*p0)[comp];

	const float oy = y - (*(const Vertex2 *)p0).y;

	(*p_intersect)[0] = oy * d[0] / d[1] + (*p0)[0];
	(*p_intersect)[1] = y;

	for (int comp = 2; comp < VecSize< VECTOR_T >(); comp++) 
		(*p_intersect)[comp] = oy * d[comp] / d[1] + (*p0)[comp];
}

// ClipByRow_ge()	: clips a polygon by a row, halfspace of greater/equal y is 'inside'
//		VECTOR_T						: one of Vector* types; must be of R2 or greater
//		- border,	const float			: halfspaces border,						input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByRow_ge(const float border,
						 const VECTOR_T *poly[],
						 uint32 &start,
						 uint32 &size,
						 VECTOR_T buffer[],
						 uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).y < border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).y > border) // is previous vertex strictly inside?
			{
				IntersectRow< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).y > border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).y < border)		// is previous vertex outside?
			{
				IntersectRow< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByCol_ge()	: clips a polygon by a column, halfspace of greater/equal x is 'inside'
//		VECTOR_T						: one of Vector* types; must be of R2 or greater
//		- border,	const float			: halfspaces border,						input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByCol_ge(const float border,
						 const VECTOR_T *poly[],
						 uint32 &start,
						 uint32 &size,
						 VECTOR_T buffer[],
						 uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).x < border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).x > border) // is previous vertex strictly inside?
			{
				IntersectCol< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).x > border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).x < border)		// is previous vertex outside?
			{
				IntersectCol< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByRow_le()	: clips a polygon by a row, halfspace of less/equal y is 'inside'
//		VECTOR_T						: one of Vector* types; must be of R2 or greater
//		- border,	const float			: halfspaces border,						input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByRow_le(const float border,
						 const VECTOR_T *poly[],
						 uint32 &start,
						 uint32 &size,
						 VECTOR_T buffer[],
						 uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).y > border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).y < border) // is previous vertex strictly inside?
			{
				IntersectRow< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).y < border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).y > border)		// is previous vertex outside?
			{
				IntersectRow< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByCol_le()	: clips a polygon by a column, halfspace of less/equal x is 'inside'
//		VECTOR_T						: one of Vector* types; must be of R2 or greater
//		- border,	const float			: halfspaces border,						input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon can be either clockwise or counter-clockwise;
//		- input pointers are not checked for validity
template < typename VECTOR_T >

inline void ClipByCol_le(const float border,
						 const VECTOR_T *poly[],
						 uint32 &start,
						 uint32 &size,
						 VECTOR_T buffer[],
						 uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		if ((*(const Vertex2 *)poly[currVert]).x > border) // is current vertex outside?
		{
			if ((*(const Vertex2 *)poly[prevVert]).x < border) // is previous vertex strictly inside?
			{
				IntersectCol< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if ((*(const Vertex2 *)poly[currVert]).x < border &&	// is current vertex strictly inside and
				(*(const Vertex2 *)poly[prevVert]).x > border)		// is previous vertex outside?
			{
				IntersectCol< VECTOR_T >(poly[prevVert], poly[currVert], border, &buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

// ClipByRect()		: clips a polygon by a screen-space rectangle
//		VECTOR_T						: one of Vector* types; must be of R2 or greater
//		- rect,		const Rect &		: clipping rectangle,						input
//		- poly,		const <VECTOR_T> **	: array containing a polygon definition,	input/output
//		- start,	uint32 &			: polygon starting vertex in the array,		input/output
//		- size,		uint32 &			: number of vertices in the polygon,		input/output
//		- buffer,	<VECTOR_T> *		: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &			: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- size would be 0 if clipping discarded all vertices
template < typename VECTOR_T >

inline void ClipByRect(const Rect &rect,
					   const VECTOR_T *poly[],
					   uint32 &start,
					   uint32 &size,
					   VECTOR_T buffer[],
					   uint32 &nbuff)
{
	ClipByRow_ge< VECTOR_T >(rect.top,	  poly, start, size, buffer, nbuff);
	ClipByCol_ge< VECTOR_T >(rect.left,   poly, start, size, buffer, nbuff);
	ClipByRow_le< VECTOR_T >(rect.bottom, poly, start, size, buffer, nbuff);
	ClipByCol_le< VECTOR_T >(rect.right,  poly, start, size, buffer, nbuff);
}

#endif // __RENDCLIPRECT__
