#include <math.h>
#include "rendSpan.h"
#include "rendClipRect.h"
#include "rendClipFrust.h"
#include "rendVect.h"
#include "rendPrim.h"
#include "rendHelper.h"

// NDCToScreen()	: maps normalized device coordinates to screen space
//		- v,		Vector3 &		: normalized device coordinates,	input/output
// returns
//		nil
// note
//		- mapping is performed accoring to the viewport definition
inline void NDCToScreen(Vector3 &v)
{
	// map the normalized device x and y coordinates to screen coordinates
	v[0] = v[0] * (viewport.screen.right - viewport.screen.left) / 2.f +
				  (viewport.screen.right + viewport.screen.left) / 2.f;

	v[1] = v[1] * (viewport.screen.top - viewport.screen.bottom) / 2.f +
				  (viewport.screen.top + viewport.screen.bottom) / 2.f;

	// map the normalized device z coordinate to the specified depth range
	v[2] = v[2] * (viewport.depth_range_f - viewport.depth_range_n) / 2.f +
				  (viewport.depth_range_f + viewport.depth_range_n) / 2.f;
}

// MapToScreen()	: maps a vector from clipping space to screen space
//		VECTOR_T						: one of Vector* types, must be of R4 or greater
//		- vi,		const <VECTOR_T> &	: vector in clipping space,	input
//		- vo,		<VECTOR_T> &		: vector in screen space,	output
// returns
//		nil
// note
//		- output vector's 4th component is the reciprocal homogeneous W
template < typename VECTOR_T >

inline void MapToScreen(const VECTOR_T &vi, VECTOR_T &vo)
{
	// perform perspective division on the clipping-space vertex,
	// normalizing the x, y and z coordinates in the course
	float oow = 1.f / vi[3];

	for (int comp = 0; comp < 3; comp++)
		vo[comp] = vi[comp] * oow;

	vo[3] = oow;

	for (int comp = 4; comp < VecSize< VECTOR_T >(); comp++)
		vo[comp] = vi[comp] * oow;

	// take the normalized device coordinates to screen space
	NDCToScreen(*(Vector3 *)&vo);
}

// Trapezoid()			: scan-converts a trapezoid polygon
//		VECTOR_T							: one of Vector* types, must be of R4 or greater
//		VERTEX_T							: one of Vertex* types with the same number of components as VECTOR_T above
//		SPAN_ROUTINE_T						: one of Span* routines which handles the above VERTEX_T type
//		- y0_p0,		const <VECTOR_T> &	: trapezoid 1st vertex,	input
//		- y0_p1,		const <VECTOR_T> &	: trapezoid 2nd vertex,	input
//		- y1_p0,		const <VECTOR_T> &	: trapezoid 3rd vertex,	input
//		- y1_p1,		const <VECTOR_T> &	: trapezoid 4th vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending-y-ascending-x order
//		- left edge is used for all height calculations
template < typename VECTOR_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &, int8 *) >

void Trapezoid(const VECTOR_T &y0_p0, const VECTOR_T &y0_p1,
			   const VECTOR_T &y1_p0, const VECTOR_T &y1_p1)
{
	// snap top end to the next greater- or equal integer;
	// snap bottom end to the previous integer
    const float snap_y0 = (float) ceilf((*(const Vertex2 *)&y0_p0).y);
    const float snap_y1 = (float) ceilf((*(const Vertex2 *)&y1_p0).y) - 1.f;

    if (snap_y1 < snap_y0)
        return;

    const float err_y0 = snap_y0 - (*(const Vertex2 *)&y0_p0).y;
    const float oody = 1.f / ((*(const Vertex2 *)&y1_p0).y - (*(const Vertex2 *)&y0_p0).y);

	// calculate the gradients of all vector components (except the Y-component) along the trapezoid's sides
	VECTOR_T dp0;
	dp0[0] = (y1_p0[0] - y0_p0[0]) * oody;

	for (int comp = 2; comp < VecSize< VECTOR_T >(); comp++)
		dp0[comp] = (y1_p0[comp] - y0_p0[comp]) * oody;

	VECTOR_T dp1;
	dp1[0] = (y1_p1[0] - y0_p1[0]) * oody;

	for (int comp = 2; comp < VecSize< VECTOR_T >(); comp++)
		dp1[comp] = (y1_p1[comp] - y0_p1[comp]) * oody;

	// iterate along the snap-corrected height-span of the trapezoid
    for (int i = 0; i < (int)(snap_y1 - snap_y0) + 1; i++)
    {
		VECTOR_T p0i;
		p0i[0] = y0_p0[0] + (err_y0 + (float) i) * dp0[0];

		for (int comp = 2; comp < VecSize< VECTOR_T >(); comp++)
			p0i[comp] = y0_p0[comp] + (err_y0 + (float) i) * dp0[comp];

		VECTOR_T p1i;
		p1i[0] = y0_p1[0] + (err_y0 + (float) i) * dp1[0];

		for (int comp = 2; comp < VecSize< VECTOR_T >(); comp++)
			p1i[comp] = y0_p1[comp] + (err_y0 + (float) i) * dp1[comp];

		SPAN_ROUTINE_T(*(VERTEX_T *)&p0i, *(VERTEX_T *)&p1i,
						renderState.frame.bits + ((int)snap_y0 + i) * renderState.frame.row_stride);
    }
}

// TriTwiceArea()	: calculates twice the signed area of a 2D triangle, as the length
//					  of the cross-product of two of the sides, taken in the XY plane
//		- p0x,		const float		: X of the triangle's 1st vertex,	input
//		- p0y,		const float		: Y of the triangle's 1st vertex,	input
//		- p1x,		const float		: X of the triangle's 2nd vertex,	input
//		- p1y,		const float		: Y of the triangle's 2nd vertex,	input
//		- p2x,		const float		: X of the triangle's 3rd vertex,	input
//		- p2y,		const float		: Y of the triangle's 3rd vertex,	input
// returns
//		float
//			negative area - counter-clockwise
//			positive area - clockwise
// note
//		- if resulting area's sign will be used as a triangle orienantion determinant,
//		  passed vertices must be in ascending y order
inline float TriTwiceArea(const float p0x, const float p0y,
						  const float p1x, const float p1y,
						  const float p2x, const float p2y)
{
	return (p1x - p0x) * (p2y - p1y) - (p2x - p1x) * (p1y - p0y);
}

// Tri()				: scan-converts a screen-space triangle, performing culling and port clipping
//		VECTOR_T							: one of Vector* types, must be of R4 or greater
//		VERTEX_T							: one of Vertex* types with the same number of components as VECTOR_T above
//		SPAN_ROUTINE_T						: one of Span* routines which handles the above VERTEX_T type
//		- p0,			const <VECTOR_T> &	: triangle 1st vertex,	input
//		- p1,			const <VECTOR_T> &	: triangle 2nd vertex,	input
//		- p2,			const <VECTOR_T> &	: triangle 3rd vertex,	input
// returns
//		nil
// note
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < typename VECTOR_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &, int8 *) >

void Tri(const VECTOR_T &p0, const VECTOR_T &p1, const VECTOR_T &p2)
{
	// find the ascending-Y order of the passed triangle vertices
	const Vertex2 *v_sorted[3];
	v_sorted[0] = (const Vertex2 *) &p0;
	v_sorted[1] = (const Vertex2 *) &p1;
	v_sorted[2] = (const Vertex2 *) &p2;

	uint32 swapped = 0;

	if ((*v_sorted[0]).y > (*v_sorted[1]).y)
	{
		Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
		swapped ^= IEEE754_SIGN_MASK;
	}
	if ((*v_sorted[1]).y > (*v_sorted[2]).y)
	{
		Swap< const Vertex2 * >(v_sorted[1], v_sorted[2]);
		swapped ^= IEEE754_SIGN_MASK;

		// vertex1 & vertex2 swapped - re-check vertex0 & vertex1
		if ((*v_sorted[0]).y > (*v_sorted[1]).y)
		{
			Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
			swapped ^= IEEE754_SIGN_MASK;
		}
	}

	// calculate the signed area of the triangle
	float triArea = TriTwiceArea((*v_sorted[0]).x, (*v_sorted[0]).y,
								 (*v_sorted[1]).x, (*v_sorted[1]).y,
								 (*v_sorted[2]).x, (*v_sorted[2]).y);

	(*(uint32*)&triArea) ^= swapped;

	// is zero area?
	const float eps = (float)1e-15;

	if ((*(uint32*)&triArea & ~IEEE754_SIGN_MASK) < *(const uint32*)&eps)
		return;

	// if culling is enabled reject the triangle if its area's sign is the same as the culling sign
	if (renderState.culling)
		if (!((*(uint32*)&triArea ^ renderState.culling) & IEEE754_SIGN_MASK))
			return;

	// clip this triangle by the viewport rectangle
	const uint32 BUFFSIZE = 40;

	const VECTOR_T *poly[BUFFSIZE];
	VECTOR_T		buff[BUFFSIZE];

	poly[0] = &p0;
	poly[1] = &p1;
	poly[2] = &p2;

	uint32 start = 0;
	uint32 size = 3;
	uint32 nbuff = 0;

	ClipByRect< VECTOR_T >(viewport.screen, poly, start, size, buff, nbuff);

	if (size == 0)
		return;

	// calculate the gradients of the vertex parameters (triangle gradients)
	float ooTriArea = 1.f / triArea;

	for (int32 g = 0; g < VecSize< VECTOR_T >() - 2; g++) // iterate the vertex parameters which we need the gradients of
	{
		renderState.trigrad_x[g] = ooTriArea * TriTwiceArea(p0[2 + g], (*(const Vertex2 *)&p0).y,
															p1[2 + g], (*(const Vertex2 *)&p1).y,
															p2[2 + g], (*(const Vertex2 *)&p2).y);

		renderState.trigrad_y[g] = ooTriArea * TriTwiceArea(p0[2 + g], (*(const Vertex2 *)&p0).x,
															p1[2 + g], (*(const Vertex2 *)&p1).x,
															p2[2 + g], (*(const Vertex2 *)&p2).x);
	}

	// find the top-most left and right, and the bottom-most left and right vertices
	int32 vert_ul = start;
	int32 vert_ur = start;
	int32 vert_ll = start;
	int32 vert_lr = start;

	for (uint32 i = start + 1; i < start + size; i++)
	{
		// check for top-most
		if ((*(const Vertex2 *)poly[i]).y < (*(const Vertex2 *)poly[vert_ul]).y) // is this vertex above the top-most?
		{
			vert_ul = i;
			vert_ur = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ul]).y) // is it at the same y as top-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ul]).x)
				vert_ul = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_ur]).x)
				vert_ur = i;
		}

		// check for bottom-most
		if ((*(const Vertex2 *)poly[i]).y > (*(const Vertex2 *)poly[vert_ll]).y) // is this vertex below the bottom-most?
		{
			vert_ll = i;
			vert_lr = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ll]).y) // is it at the same y as bottom-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ll]).x)
				vert_ll = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_lr]).x)
				vert_lr = i;
		}
	}

	// divide the polygon vertically into non-overlapping trapezoids
	int32 inc_l;

	if (*(uint32 *)&triArea & IEEE754_SIGN_MASK)	// is poly counter-clockwise?
		inc_l = 1;
	else											// poly is clockwise
		inc_l = -1;

	int32 next_l = start + (size + (vert_ul - start) + inc_l) % size;
	int32 next_r = start + (size + (vert_ur - start) - inc_l) % size;

	const VECTOR_T *top_l = poly[vert_ul];
	const VECTOR_T *top_r = poly[vert_ur];

	VECTOR_T *tmp = buff + nbuff;

	while (vert_ul != vert_ll && vert_ur != vert_lr)
	{
		if ((*(const Vertex2 *)poly[next_l]).y == (*(const Vertex2 *)poly[next_r]).y) // do both segments end simultaneously?
		{
			Trapezoid< VECTOR_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															*poly[next_l], *poly[next_r]);
			// symetrical advancement
			vert_ul = next_l;
			vert_ur = next_r;

			next_l = start + (size + (next_l - start) + inc_l) % size;
			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = poly[vert_ul];
			top_r = poly[vert_ur];
		}
		else
		if ((*(const Vertex2 *)poly[next_l]).y < (*(const Vertex2 *)poly[next_r]).y) // left segment ends first
		{
			IntersectRow< VECTOR_T >(poly[vert_ur], poly[next_r], (*(const Vertex2 *)poly[next_l]).y, tmp);

			Trapezoid< VECTOR_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															*poly[next_l], *tmp);
			// advance left
			vert_ul = next_l;

			next_l = start + (size + (next_l - start) + inc_l) % size;

			top_l = poly[vert_ul];
			top_r = tmp++;
		}
		else // right segment ends first
		{
			IntersectRow< VECTOR_T >(poly[vert_ul], poly[next_l], (*(const Vertex2 *)poly[next_r]).y, tmp);

			Trapezoid< VECTOR_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															*tmp, *poly[next_r]);
			// advance right
			vert_ur = next_r;

			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = tmp++;
			top_r = poly[vert_ur];
		}
	}
}

// Triangle()			: draws a triangle, taking it through a model-projection transform,
//						  frustum-clipping it, culling it and port-clipping it
//		VECTOR_T							: one of Vector* types, must be of R4 or greater
//		VERTEX_T							: one of Vertex* types with the same number of components as VECTOR_T above
//		SPAN_ROUTINE_T						: one of Span* routines which handles the above VERTEX_T type
//		- mat,			const Matrix4 &		: model-projection matrix,	input
//		- p0,			const <VECTOR_T> &	: triangle 1st vertex,		input
//		- p1,			const <VECTOR_T> &	: triangle 2nd vertex,		input
//		- p2,			const <VECTOR_T> &	: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < typename VECTOR_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &, int8 *) >

void Triangle(const Matrix4 &mat, const VECTOR_T &p0, const VECTOR_T &p1, const VECTOR_T &p2)
{
	const uint32 BUFFSIZE = 40;

	VECTOR_T p_clip[BUFFSIZE];

	Transform< VECTOR_T >(mat, p0, p_clip[0]);
	Transform< VECTOR_T >(mat, p1, p_clip[1]);
	Transform< VECTOR_T >(mat, p2, p_clip[2]);

	const VECTOR_T *poly[BUFFSIZE];

	poly[0] = p_clip + 0;
	poly[1] = p_clip + 1;
	poly[2] = p_clip + 2;

	uint32 start = 0;
	uint32 size = 3;
	uint32 nclip = 3;

	ClipByFrustum< VECTOR_T >(poly, start, size, p_clip, nclip);

	if (size == 0)
		return;

	VECTOR_T p_scr[BUFFSIZE];

	for (uint32 i = 0; i < size; i++)
		MapToScreen< VECTOR_T >(*poly[start + i], p_scr[i]);

	// rasterize the resulting polygon as a triangle fan
	for (uint32 i = 0; i < size - 2; i++)
		Tri< VECTOR_T, VERTEX_T, SPAN_ROUTINE_T >(p_scr[0], p_scr[i + 1], p_scr[i + 2]);
}

// PlotTriangle6()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const Matrix4 &		: model-projection matrix,	input
//		- p0,	const Vector6 &		: triangle 1st vertex,		input
//		- p1,	const Vector6 &		: triangle 2nd vertex,		input
//		- p2,	const Vector6 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(const Matrix4 &mat, const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	switch((renderState.sampling << 24) | (renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)	   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:

		Triangle< Vector6, Vertex4UV, Span4UV_any_nop_any >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_linear_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_linear_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_linear_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_linear_cpy_rgb888 >(mat, p0, p1, p2);
		break;
		
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_rgb565 >(mat, p0, p1, p2);
		break;
	}
}

// PlotTriangle6()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const Vector6 &		: triangle 1st vertex,		input
//		- p1,	const Vector6 &		: triangle 2nd vertex,		input
//		- p2,	const Vector6 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(const Vector6 &p0, const Vector6 &p1, const Vector6 &p2)
{
	switch((renderState.sampling << 24) | (renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_NEAREST << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)    | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_CONST_LINEAR << 24)	   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_NEAREST << 24)  | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_SAMP_LINEAR_LINEAR << 24)   | (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_NOP:

		Tri< Vector6, Vertex4UV, Span4UV_any_nop_any >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_linear_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_linear_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_linear_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_CONST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_linear_cpy_rgb888 >(p0, p1, p2);
		break;
		
	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_NEAREST << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_nearest_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_SAMP_NEAREST_LINEAR << 24) | (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_UV << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector6, Vertex4UV, Span4UV_nearest_linear_cpy_rgb565 >(p0, p1, p2);
		break;
	}
}

// PlotTriangle8()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const Matrix4 &		: model-projection matrix,	input
//		- p0,	const Vector8 &		: triangle 1st vertex,		input
//		- p1,	const Vector8 &		: triangle 2nd vertex,		input
//		- p2,	const Vector8 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle8(const Matrix4 &mat, const Vector8 &p0, const Vector8 &p1, const Vector8 &p2)
{
	switch((renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_nop_any >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_cpy_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_cpy_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_cpy_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_cpy_rgb888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_mod_argb1555 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_mod_argb8888 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_mod_rgb565 >(mat, p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Triangle< Vector8, Vertex4ARGB, Span4ARGB_mod_rgb888 >(mat, p0, p1, p2);
		break;
	}
}

// PlotTriangle8()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const Vector8 &		: triangle 1st vertex,		input
//		- p1,	const Vector8 &		: triangle 2nd vertex,		input
//		- p2,	const Vector8 &		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle8(const Vector8 &p0, const Vector8 &p1, const Vector8 &p2)
{
	switch((renderState.frame.pix_format << 16) | (renderState.vert_format << 8) | renderState.color_op)
	{
	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_RGB565 << 16)   | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:
	case (RENDER_STATE_PIX_RGB888 << 16)   | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_NOP:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_nop_any >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_cpy_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_cpy_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_cpy_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_COPY:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_cpy_rgb888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB1555 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_mod_argb1555 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_ARGB8888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_mod_argb8888 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB565 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_mod_rgb565 >(p0, p1, p2);
		break;

	case (RENDER_STATE_PIX_RGB888 << 16) | (RENDER_STATE_VERT_ARGB << 8) | RENDER_STATE_OP_MODULATE:

		Tri< Vector8, Vertex4ARGB, Span4ARGB_mod_rgb888 >(p0, p1, p2);
		break;
	}
}
