#include <math.h>
#include <mathinline.h>
#include "rendSpan.h"
#include "rendHelper.h"

// Span4UV_any_nop_any()	: scan-converts a horizontal span into the void
//		- p0,				const Vertex4UV &	: span 1st vertex,	input
//		- p1,				const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,		int8 *				: output scanline,	input
// returns
//		nil
void Span4UV_any_nop_any(const Vertex4UV &p0,
						 const Vertex4UV &p1,
						 int8 *linebuffer)
{ }

// Span4UV_nearest_cpy_argb8888()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_argb8888(const Vertex4UV &p0,
								  const Vertex4UV &p1,
								  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = tex;
    }
}

// Span4UV_nearest_cpy_argb1555()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_argb1555(const Vertex4UV &p0,
								  const Vertex4UV &p1,
								  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelARGB1555 formatted;

		formatted.a = (uint8) tex.a >> 7;
		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 3;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = formatted;
    }
}

// Span4UV_nearest_cpy_rgb888()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_rgb888(const Vertex4UV &p0,
								const Vertex4UV &p1,
								int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB888 formatted;

		formatted.r = (uint8) tex.r;
		formatted.g = (uint8) tex.g;
		formatted.b = (uint8) tex.b;

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = formatted;
    }
}

// Span4UV_nearest_cpy_rgb565()	: scan-converts a horizontal span, featuring nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_cpy_rgb565(const Vertex4UV &p0,
								const Vertex4UV &p1,
								int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

        snap_ui %= renderState.tex.dimx;
        snap_vi %= renderState.tex.dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														snap_vi * renderState.tex.row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB565 formatted;

		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 2;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = formatted;
    }
}

// Span4UV_linear_cpy_argb8888()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_argb8888(const Vertex4UV &p0,
								 const Vertex4UV &p1,
								 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 filtered;

		filtered.a = (uint8) floorf(.5f +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11);
		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = filtered;
    }
}

// Span4UV_linear_cpy_argb1555()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,						const Vertex4UV &	: span 1st vertex,	input
//		- p1,						const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,				int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_argb1555(const Vertex4UV &p0,
								 const Vertex4UV &p1,
								 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB1555 filtered;

		filtered.a = (uint8) floorf(.5f +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11) >> 7;
		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 3;
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = filtered;
    }
}

// Span4UV_linear_cpy_rgb888()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_rgb888(const Vertex4UV &p0,
							   const Vertex4UV &p1,
							   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB888 filtered;

		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = filtered;
    }
}

// Span4UV_linear_cpy_rgb565()	: scan-converts a horizontal span, featuring linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_linear_cpy_rgb565(const Vertex4UV &p0,
							   const Vertex4UV &p1,
							   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float duow_dx = renderState.trigrad_x[2];
    const float dvow_dx = renderState.trigrad_x[3];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex.dimx;
        vi0 %= renderState.tex.dimy;
        ui1 %= renderState.tex.dimx;
        vi1 %= renderState.tex.dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi0 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex.bits_const +
														vi1 * renderState.tex.row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB565 filtered;

		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 2;
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = filtered;
    }
}

// Span4UV_nearest_nearest_cpy_argb8888()	: scan-converts a horizontal span, featuring nearest mipmap selection
//											  and nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_nearest_cpy_argb8888(const Vertex4UV &p0,
										  const Vertex4UV &p1,
										  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

		snap_ui %= renderState.tex_lod[lod].dimx;
		snap_vi %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														snap_vi * renderState.tex_lod[lod].row_stride +
														snap_ui * sizeof(PixelARGB8888));

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = tex;
	}
}

// Span4UV_nearest_nearest_cpy_argb1555()	: scan-converts a horizontal span, featuring nearest mipmap selection
//											  and nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_nearest_cpy_argb1555(const Vertex4UV &p0,
										  const Vertex4UV &p1,
										  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

		snap_ui %= renderState.tex_lod[lod].dimx;
		snap_vi %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														snap_vi * renderState.tex_lod[lod].row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelARGB1555 formatted;

		formatted.a = (uint8) tex.a >> 7;
		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 3;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = formatted;
	}
}

// Span4UV_nearest_nearest_cpy_rgb888()	: scan-converts a horizontal span, featuring nearest mipmap selection
//										  and nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_nearest_cpy_rgb888(const Vertex4UV &p0,
										const Vertex4UV &p1,
										int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

		snap_ui %= renderState.tex_lod[lod].dimx;
		snap_vi %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														snap_vi * renderState.tex_lod[lod].row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB888 formatted;

		formatted.r = (uint8) tex.r;
		formatted.g = (uint8) tex.g;
		formatted.b = (uint8) tex.b;

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = formatted;
	}
}

// Span4UV_nearest_nearest_cpy_rgb565()	: scan-converts a horizontal span, featuring nearest mipmap selection
//										  and nearest texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_nearest_cpy_rgb565(const Vertex4UV &p0,
										const Vertex4UV &p1,
										int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		int snap_ui = (int) floorf(ui + .5f);
		int snap_vi = (int) floorf(vi + .5f);

		snap_ui %= renderState.tex_lod[lod].dimx;
		snap_vi %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														snap_vi * renderState.tex_lod[lod].row_stride +
														snap_ui * sizeof(PixelARGB8888));
		PixelRGB565 formatted;

		formatted.r = (uint8) tex.r >> 3;
		formatted.g = (uint8) tex.g >> 2;
		formatted.b = (uint8) tex.b >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = formatted;
	}
}

// Span4UV_nearest_linear_cpy_argb8888()	: scan-converts a horizontal span, featuring nearest mipmap selection
//											  and linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_linear_cpy_argb8888(const Vertex4UV &p0,
										 const Vertex4UV &p1,
										 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex_lod[lod].dimx;
        vi0 %= renderState.tex_lod[lod].dimy;
        ui1 %= renderState.tex_lod[lod].dimx;
        vi1 %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 filtered;

		filtered.a = (uint8) floorf(.5f +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11);
		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = filtered;
	}
}

// Span4UV_nearest_linear_cpy_argb1555()	: scan-converts a horizontal span, featuring nearest mipmap selection
//											  and linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_linear_cpy_argb1555(const Vertex4UV &p0,
										 const Vertex4UV &p1,
										 int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex_lod[lod].dimx;
        vi0 %= renderState.tex_lod[lod].dimy;
        ui1 %= renderState.tex_lod[lod].dimx;
        vi1 %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB1555 filtered;

		filtered.a = (uint8) floorf(.5f +
			(float)tex00.a * mass00 + (float)tex01.a * mass01 + (float)tex10.a * mass10 + (float)tex11.a * mass11) >> 7;
		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 3;
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = filtered;
	}
}

// Span4UV_nearest_linear_cpy_rgb888()	: scan-converts a horizontal span, featuring nearest mipmap selection
//										  and linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_linear_cpy_rgb888(const Vertex4UV &p0,
									   const Vertex4UV &p1,
									   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex_lod[lod].dimx;
        vi0 %= renderState.tex_lod[lod].dimy;
        ui1 %= renderState.tex_lod[lod].dimx;
        vi1 %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB888 filtered;

		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11);
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11);
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11);

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = filtered;
	}
}

// Span4UV_nearest_linear_cpy_rgb565()	: scan-converts a horizontal span, featuring nearest mipmap selection
//										  and linear texture sampling, op. COPY
//		- p0,					const Vertex4UV &	: span 1st vertex,	input
//		- p1,					const Vertex4UV &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4UV_nearest_linear_cpy_rgb565(const Vertex4UV &p0,
									   const Vertex4UV &p1,
									   int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0 = snap_x0 - p0.x;

	const float dz_dx	= renderState.trigrad_x[0];
	const float doow_dx = renderState.trigrad_x[1];
	const float duow_dx = renderState.trigrad_x[2];
	const float dvow_dx = renderState.trigrad_x[3];

	const float doow_dy = renderState.trigrad_y[1];
	const float duow_dy = renderState.trigrad_y[2];
	const float dvow_dy = renderState.trigrad_y[3];

	// calculate the uv-mapping of a fictitious pixel on the left of the span
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;

	float uiLeft = uow * w;
	float viLeft = vow * w;

	const int span_len = (int)(snap_x1 - snap_x0) + 1;
	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		// calculate the uv-mapping of a fictitious pixel above this one
		const float oowiUp = oowi - doow_dy;
		const float uowiUp = uowi - duow_dy;
		const float vowiUp = vowi - dvow_dy;

		const float wiUp = 1.f / oowiUp;
		const float uiUp = uowiUp * wiUp;
		const float viUp = vowiUp * wiUp;

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		float dudx = ui - uiLeft;
		float dvdx = vi - viLeft;
		float dudy = ui - uiUp;
		float dvdy = vi - viUp;
		(*(uint32*)&dudx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdx) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dudy) &= ~IEEE754_SIGN_MASK;
		(*(uint32*)&dvdy) &= ~IEEE754_SIGN_MASK;

		uiLeft = ui;
		viLeft = vi;

		// calculate the texture scale factor (>1 is minification)
		const float duMax = Max(dudx, dudy);
		const float dvMax = Max(dvdx, dvdy);
		const float rho = Max(duMax, dvMax);

		// pick a texture LOD based on the scale factor
		int lod = (int) floorf(log2(rho) + .5f);

//		const float oolog2 = 1.f / logf(2);
//		int lod = (int) floorf(logf(rho) * oolog2 + .5f);
		/* use the above if log2(x) is not available; utilizes the property that logN(x) = log(x) / log(N) */
		
		if (lod < renderState.tex_min_lod)
			lod = renderState.tex_min_lod;
		else
		if (lod > renderState.tex_max_lod)
			lod = renderState.tex_max_lod;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		const float snap_ui = floorf(ui);
		const float snap_vi = floorf(vi);

        const float mass_u1 = ui - snap_ui;
        const float mass_v1 = vi - snap_vi;

		const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
		const float mass01 = mass_u1		 * (1.f - mass_v1);
		const float mass11 = mass_u1		 * mass_v1;
		const float mass10 = (1.f - mass_u1) * mass_v1;

        int ui0 = (int) snap_ui;
        int vi0 = (int) snap_vi;
        int ui1 = ui0 + 1;
        int vi1 = vi0 + 1;

        ui0 %= renderState.tex_lod[lod].dimx;
        vi0 %= renderState.tex_lod[lod].dimy;
        ui1 %= renderState.tex_lod[lod].dimx;
        vi1 %= renderState.tex_lod[lod].dimy;

		PixelARGB8888 tex00 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex01 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi0 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelARGB8888 tex10 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui0 * sizeof(PixelARGB8888));
		PixelARGB8888 tex11 = *(const PixelARGB8888 *)(renderState.tex_lod[lod].bits_const +
														vi1 * renderState.tex_lod[lod].row_stride +
														ui1 * sizeof(PixelARGB8888));
		PixelRGB565 filtered;

		filtered.r = (uint8) floorf(.5f +
			(float)tex00.r * mass00 + (float)tex01.r * mass01 + (float)tex10.r * mass10 + (float)tex11.r * mass11) >> 3;
		filtered.g = (uint8) floorf(.5f +
			(float)tex00.g * mass00 + (float)tex01.g * mass01 + (float)tex10.g * mass10 + (float)tex11.g * mass11) >> 2;
		filtered.b = (uint8) floorf(.5f +
			(float)tex00.b * mass00 + (float)tex01.b * mass01 + (float)tex10.b * mass10 + (float)tex11.b * mass11) >> 3;

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = filtered;
	}
}

// Span4ARGB_nop_any()		: scan-converts a horizontal span into the void
//		- p0,				const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,				const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,		int8 *				: output scanline,	input
// returns
//		nil
void Span4ARGB_nop_any(const Vertex4ARGB &,
					   const Vertex4ARGB &,
					   int8 *)
{ }

// Span4ARGB_cpy_argb8888()		: scan-converts a horizontal span, op. COPY
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_cpy_argb8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1,
							int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelARGB8888 formatted;
		formatted.a = (uint8) floorf(.5f + 255.f * ai);
		formatted.r = (uint8) floorf(.5f + 255.f * ri);
		formatted.g = (uint8) floorf(.5f + 255.f * gi);
		formatted.b = (uint8) floorf(.5f + 255.f * bi);

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = formatted;
    }
}

// Span4ARGB_cpy_argb1555()		: scan-converts a horizontal span, op. COPY
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_cpy_argb1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1,
							int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelARGB1555 formatted;
		formatted.a = (uint8) floorf(.5f +	1.f * ai);
		formatted.r = (uint8) floorf(.5f + 31.f * ri);
		formatted.g = (uint8) floorf(.5f + 31.f * gi);
		formatted.b = (uint8) floorf(.5f + 31.f * bi);

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = formatted;
    }
}

// Span4ARGB_cpy_rgb888()		: scan-converts a horizontal span, op. COPY
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_cpy_rgb888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1,
						  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelRGB888 formatted;
		formatted.r = (uint8) floorf(.5f + 255.f * ri);
		formatted.g = (uint8) floorf(.5f + 255.f * gi);
		formatted.b = (uint8) floorf(.5f + 255.f * bi);

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = formatted;
    }
}

// Span4ARGB_cpy_rgb565()		: scan-converts a horizontal span, op. COPY
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_cpy_rgb565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1,
						  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelRGB565 formatted;
		formatted.r = (uint8) floorf(.5f + 31.f * ri);
		formatted.g = (uint8) floorf(.5f + 63.f * gi);
		formatted.b = (uint8) floorf(.5f + 31.f * bi);

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = formatted;
    }
}

// Span4ARGB_mod_argb8888()		: scan-converts a horizontal span, op. MODULATE
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_mod_argb8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1,
							int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelARGB8888 multiplicand = *(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888));

		PixelARGB8888 formatted;
		formatted.a = (uint8) floorf(.5f + (float) multiplicand.a * ai);
		formatted.r = (uint8) floorf(.5f + (float) multiplicand.r * ri);
		formatted.g = (uint8) floorf(.5f + (float) multiplicand.g * gi);
		formatted.b = (uint8) floorf(.5f + (float) multiplicand.b * bi);

		*(PixelARGB8888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB8888)) = formatted;
    }
}

// Span4ARGB_mod_argb1555()		: scan-converts a horizontal span, op. MODULATE
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_mod_argb1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1,
							int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelARGB1555 multiplicand = *(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555));

		PixelARGB1555 formatted;
		formatted.a = (uint8) floorf(.5f + (float) multiplicand.a * ai);
		formatted.r = (uint8) floorf(.5f + (float) multiplicand.r * ri);
		formatted.g = (uint8) floorf(.5f + (float) multiplicand.g * gi);
		formatted.b = (uint8) floorf(.5f + (float) multiplicand.b * bi);

		*(PixelARGB1555 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelARGB1555)) = formatted;
    }
}

// Span4ARGB_mod_rgb888()		: scan-converts a horizontal span, op. MODULATE
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_mod_rgb888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1,
						  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelRGB888 multiplicand = *(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888));

		PixelRGB888 formatted;
		formatted.r = (uint8) floorf(.5f + (float) multiplicand.r * ri);
		formatted.g = (uint8) floorf(.5f + (float) multiplicand.g * gi);
		formatted.b = (uint8) floorf(.5f + (float) multiplicand.b * bi);

		*(PixelRGB888 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB888)) = formatted;
    }
}

// Span4ARGB_mod_rgb565()		: scan-converts a horizontal span, op. MODULATE
//		- p0,					const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,					const Vertex4ARGB &	: span 2nd vertex,	input
//		- linebuffer,			int8 *				: output scanline,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order
void Span4ARGB_mod_rgb565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1,
						  int8 *linebuffer)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0 = snap_x0 - p0.x;

    const float dz_dx	= renderState.trigrad_x[0];
    const float doow_dx = renderState.trigrad_x[1];
    const float daow_dx = renderState.trigrad_x[2];
    const float drow_dx = renderState.trigrad_x[3];
    const float dgow_dx = renderState.trigrad_x[4];
    const float dbow_dx = renderState.trigrad_x[5];

    const int span_len = (int)(snap_x1 - snap_x0) + 1;
    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
        const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
        const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
        const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

        const float ai = aowi * wi;
        const float ri = rowi * wi;
        const float gi = gowi * wi;
        const float bi = bowi * wi;

		PixelRGB565 multiplicand = *(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565));

		PixelRGB565 formatted;
		formatted.r = (uint8) floorf(.5f + (float) multiplicand.r * ri);
		formatted.g = (uint8) floorf(.5f + (float) multiplicand.g * gi);
		formatted.b = (uint8) floorf(.5f + (float) multiplicand.b * bi);

		*(PixelRGB565 *)(linebuffer + ((int)snap_x0 + i) * sizeof(PixelRGB565)) = formatted;
    }
}
