#include <math.h>
#include "rendVect.h"

void Rotate(Matrix4 &mat, float a, float x, float y, float z)
{
	Vector4 v = { x, y, z, 0 };
	MulVect4(v, 1.f / LenVect4(v));

	const float sin_a = sinf(a);
	const float cos_a = cosf(a);

	x = v[0];
	y = v[1];
	z = v[2];
	
	SetMat4(mat,
			x * x + cos_a * (1 - x * x),		 y * x - cos_a * (y * x) - sin_a * z, z * x - cos_a * (z * x) + sin_a * y, 0,
			x * y - cos_a * (x * y) + sin_a * z, y * y + cos_a * (1 - y * y),		  z * y - cos_a * (z * y) - sin_a * x, 0,
			x * z - cos_a * (x * z) - sin_a * y, y * z - cos_a * (y * z) + sin_a * x, z * z + cos_a * (1 - z * z),		   0,
			0,									 0,									  0,								   1);
}

void Translate(Matrix4 &mat, float x, float y, float z)
{
	SetMat4(mat,
			1,	0,	0,	x,
			0,	1,	0,	y,
			0,	0,	1,	z,
			0,	0,	0,	1);
}

void Scale(Matrix4 &mat, float x, float y, float z)
{
	SetMat4(mat,
			x,	0,	0,	0,
			0,	y,	0,	0,
			0,	0,	z,	0,
			0,	0,	0,	1);
}

// Frustum()	: builds a perspective-projection transform
//		- mat,	Matrix4 &	: matrix of the projection transform,								output
//		- l,	float		: left edge of viewport's mapping onto the near clipping plane,		input
//		- r,	float		: right edge of viewport's mapping onto the near clipping plane,	input
//		- b,	float		: bottom edge of viewport's mapping onto the near clipping plane,	input
//		- t,	float		: top edge of viewport's mapping onto the near clipping plane,		input
//		- n,	float		: absolute distance from the origin to the near clipping plane,		input
//		- f,	float		: absolute distance from the origin to the far clipping plane,		input
// returns
//		nil
// note
//		- camera space taken as right-handed, positive y-axis pointing up
//		  and negative z-axis pointing in the direction of the view vector
//		- the distances to the near and far clipping planes are in absolute values,
//		  whereas the two planes are at -n and -f units from the origin, respectively
//		- clipping is	-Wc <= Xc <= Wc,	and DNC is	-1 <= Xd <= 1
//						-Wc <= Yc <= Wc					-1 <= Yd <= 1
//						-Wc <= Zc <= Wc					-1 <= Zd <= 1
//		- projection plane equals the near clipping plane
void Frustum(Matrix4 &mat, float l, float r, float b, float t, float n, float f)
{
	SetMat4(mat,
			2 * n / (r - l),	0,					(r + l) / (r - l),	0,
			0,					2 * n / (t - b),	(t + b) / (t - b),	0,
			0,					0,				  - (f + n) / (f - n),	- 2 * f * n / (f - n),
			0,					0,				  - 1,					0);
}

// Ortho()		: builds a parallel-projection transform
//		- mat,	Matrix4 &	: matrix of the projection transform,								output
//		- l,	float		: left edge of viewport's mapping onto the near clipping plane,		input
//		- r,	float		: right edge of viewport's mapping onto the near clipping plane,	input
//		- b,	float		: bottom edge of viewport's mapping onto the near clipping plane,	input
//		- t,	float		: top edge of viewport's mapping onto the near clipping plane,		input
//		- n,	float		: absolute distance from the origin to the near clipping plane,		input
//		- f,	float		: absolute distance from the origin to the far clipping plane,		input
// returns
//		nil
// note
//		- camera space taken as right-handed, positive y-axis pointing up
//		  and negative z-axis pointing in the direction of the view vector
//		- the distances to the near and far clipping planes are in absolute values,
//		  whereas the two planes are at -n and -f units from the origin, respectively
//		- clipping is	-Wc <= Xc <= Wc,	and DNC is	-1 <= Xd <= 1
//						-Wc <= Yc <= Wc					-1 <= Yd <= 1
//						-Wc <= Zc <= Wc					-1 <= Zd <= 1
//		- projection plane equals the near clipping plane
void Ortho(Matrix4 &mat, float l, float r, float b, float t, float n, float f)
{
	SetMat4(mat,
			2 / (r - l),	0,				0,				- (r + l) / (r - l),
			0,				2 / (t - b), 	0,				- (t + b) / (t - b),
			0,				0,				- 2 / (f - n),	- (f + n) / (f - n),
			0,				0,				0,				1 );
}

// InvertMat4()		: finds the inverse of the given matrix, given latter is invertible
//		- mat,		const Matrix4 &	: argument matrix,								input
//		- inv,		Matrix4 &		: inverse of the argument,						output
// returns
//		bool
// note
//		- return value: true - succeeded, false - matrix is non-invertible;
//		- uses Cramer's rule for matrix inversion.
bool InvertMat4(const Matrix4 &mat, Matrix4 &inv)
{
	float *dst = (float *)&inv;
	float  src[16], tmp[12];
	int    i;

	// transpose source matrix
	for ( i = 0; i < 4; i++ )
	{
		src[i +  0] = mat[i][0];
		src[i +  4] = mat[i][1];
		src[i +  8] = mat[i][2];
		src[i + 12] = mat[i][3];
	}

	// calculate pairs for first 8 elements (cofactors)
	tmp[ 0]  = src[10] * src[15];
	tmp[ 1]  = src[11] * src[14];
	tmp[ 2]  = src[ 9] * src[15];
	tmp[ 3]  = src[11] * src[13];
	tmp[ 4]  = src[ 9] * src[14];
	tmp[ 5]  = src[10] * src[13];
	tmp[ 6]  = src[ 8] * src[15];
	tmp[ 7]  = src[11] * src[12];
	tmp[ 8]  = src[ 8] * src[14];
	tmp[ 9]  = src[10] * src[12];
	tmp[10]  = src[ 8] * src[13];
	tmp[11]  = src[ 9] * src[12];

	// calculate first 8 elements (cofactors)
	dst[ 0]  = tmp[ 0] * src[ 5] + tmp[ 3] * src[ 6] + tmp[ 4] * src[ 7];
	dst[ 0] -= tmp[ 1] * src[ 5] + tmp[ 2] * src[ 6] + tmp[ 5] * src[ 7];
	dst[ 1]  = tmp[ 1] * src[ 4] + tmp[ 6] * src[ 6] + tmp[ 9] * src[ 7];
	dst[ 1] -= tmp[ 0] * src[ 4] + tmp[ 7] * src[ 6] + tmp[ 8] * src[ 7];
	dst[ 2]  = tmp[ 2] * src[ 4] + tmp[ 7] * src[ 5] + tmp[10] * src[ 7];
	dst[ 2] -= tmp[ 3] * src[ 4] + tmp[ 6] * src[ 5] + tmp[11] * src[ 7];
	dst[ 3]  = tmp[ 5] * src[ 4] + tmp[ 8] * src[ 5] + tmp[11] * src[ 6];
	dst[ 3] -= tmp[ 4] * src[ 4] + tmp[ 9] * src[ 5] + tmp[10] * src[ 6];
	dst[ 4]  = tmp[ 1] * src[ 1] + tmp[ 2] * src[ 2] + tmp[ 5] * src[ 3];
	dst[ 4] -= tmp[ 0] * src[ 1] + tmp[ 3] * src[ 2] + tmp[ 4] * src[ 3];
	dst[ 5]  = tmp[ 0] * src[ 0] + tmp[ 7] * src[ 2] + tmp[ 8] * src[ 3];
	dst[ 5] -= tmp[ 1] * src[ 0] + tmp[ 6] * src[ 2] + tmp[ 9] * src[ 3];
	dst[ 6]  = tmp[ 3] * src[ 0] + tmp[ 6] * src[ 1] + tmp[11] * src[ 3];
	dst[ 6] -= tmp[ 2] * src[ 0] + tmp[ 7] * src[ 1] + tmp[10] * src[ 3];
	dst[ 7]  = tmp[ 4] * src[ 0] + tmp[ 9] * src[ 1] + tmp[10] * src[ 2];
	dst[ 7] -= tmp[ 5] * src[ 0] + tmp[ 8] * src[ 1] + tmp[11] * src[ 2];

	// calculate pairs for second 8 elements (cofactors)
	tmp[ 0]  = src[ 2] * src[ 7];
	tmp[ 1]  = src[ 3] * src[ 6];
	tmp[ 2]  = src[ 1] * src[ 7];
	tmp[ 3]  = src[ 3] * src[ 5];
	tmp[ 4]  = src[ 1] * src[ 6];
	tmp[ 5]  = src[ 2] * src[ 5];
	tmp[ 6]  = src[ 0] * src[ 7];
	tmp[ 7]  = src[ 3] * src[ 4];
	tmp[ 8]  = src[ 0] * src[ 6];
	tmp[ 9]  = src[ 2] * src[ 4];
	tmp[10]  = src[ 0] * src[ 5];
	tmp[11]  = src[ 1] * src[ 4];

	// calculate second 8 elements (cofactors)
	dst[ 8]  = tmp[ 0] * src[13] + tmp[ 3] * src[14] + tmp[ 4] * src[15];
	dst[ 8] -= tmp[ 1] * src[13] + tmp[ 2] * src[14] + tmp[ 5] * src[15];
	dst[ 9]  = tmp[ 1] * src[12] + tmp[ 6] * src[14] + tmp[ 9] * src[15];
	dst[ 9] -= tmp[ 0] * src[12] + tmp[ 7] * src[14] + tmp[ 8] * src[15];
	dst[10]  = tmp[ 2] * src[12] + tmp[ 7] * src[13] + tmp[10] * src[15];
	dst[10] -= tmp[ 3] * src[12] + tmp[ 6] * src[13] + tmp[11] * src[15];
	dst[11]  = tmp[ 5] * src[12] + tmp[ 8] * src[13] + tmp[11] * src[14];
	dst[11] -= tmp[ 4] * src[12] + tmp[ 9] * src[13] + tmp[10] * src[14];
	dst[12]  = tmp[ 2] * src[10] + tmp[ 5] * src[11] + tmp[ 1] * src[ 9];
	dst[12] -= tmp[ 4] * src[11] + tmp[ 0] * src[ 9] + tmp[ 3] * src[10];
	dst[13]  = tmp[ 8] * src[11] + tmp[ 0] * src[ 8] + tmp[ 7] * src[10];
	dst[13] -= tmp[ 6] * src[10] + tmp[ 9] * src[11] + tmp[ 1] * src[ 8];
	dst[14]  = tmp[ 6] * src[ 9] + tmp[11] * src[11] + tmp[ 3] * src[ 8];
	dst[14] -= tmp[10] * src[11] + tmp[ 2] * src[ 8] + tmp[ 7] * src[ 9];
	dst[15]  = tmp[10] * src[10] + tmp[ 4] * src[ 8] + tmp[ 9] * src[ 9];
	dst[15] -= tmp[ 8] * src[ 9] + tmp[11] * src[10] + tmp[ 5] * src[ 8];

	// calculate reciprocal determinant
	const float eps = (float)1e-15;
	float det = src[0] * dst[0] + src[1] * dst[1] + src[2] * dst[2] + src[3] * dst[3];

	if ( fabs( det ) < eps )
		return false;

	// calculate matrix inverse
	det = 1.f / det;
	for ( i = 0; i < 16; i++ )
		dst[i] *= det;

	return true;
}
